
module cwx.cwscviewer;

import dfl.all;

import core.thread;

import std.algorithm;
import std.file;
import std.path;
import std.process;
import std.stdio;
import std.stream;
import std.string;
import std.utf;

import cwx.cab;
import cwx.cwls;
import cwx.sjis;

private alias std.file.copy fcopy;
private alias std.file.remove fremove;
private alias std.path.join pjoin;

private version (Windows) {
	const NO_FN_C = "\\/:*?\"<>|"d;
	const NO_FN_CC = "￥／：＊？”＜＞｜"d;
	static assert (NO_FN_C.length == NO_FN_CC.length);
	import std.c.stdio;
	import std.c.stdlib;
	extern (Windows) {
		import std.c.windows.windows;
		struct STARTUPINFO {
			DWORD cb = 0;
			LPTSTR lpReserved = null;
			LPTSTR lpDesktop = null;
			LPTSTR lpTitle = null;
			DWORD dwX = 0;
			DWORD dwY = 0;
			DWORD dwXSize = 0;
			DWORD dwYSize = 0;
			DWORD dwXCountChars = 0;
			DWORD dwYCountChars = 0;
			DWORD dwFillAttribute = 0;
			DWORD dwFlags = 0;
			USHORT wShowWindow = 0;
			USHORT cbReserved2 = 0;
			LPBYTE lpReserved2 = null;
			HANDLE hStdInput = null;
			HANDLE hStdOutput = null;
			HANDLE hStdError = null;
		}
		struct PROCESS_INFORMATION {
			HANDLE hProcess;
			HANDLE hThread;
			DWORD dwProcessId;
			DWORD dwThreadId;
		}
		BOOL SetFileAttributesA(LPSTR, DWORD);
		BOOL SetFileAttributesW(LPWSTR, DWORD);
		BOOL CreateProcessA(LPCSTR, LPSTR, SECURITY_ATTRIBUTES*, SECURITY_ATTRIBUTES*,
			BOOL, DWORD, LPVOID, LPCSTR, STARTUPINFO*, PROCESS_INFORMATION*);
		BOOL CreateProcessW(LPCWSTR, LPWSTR, SECURITY_ATTRIBUTES*, SECURITY_ATTRIBUTES*,
			BOOL, DWORD, LPVOID, LPCWSTR, STARTUPINFO*, PROCESS_INFORMATION*);
	}
	/// processを実行する。workDirを作業フォルダとする。
	bool exec(string process, string workDir = "") {
		STARTUPINFO setup;
		setup.cb = setup.sizeof;
		PROCESS_INFORMATION info;

		int r;
		if (GetVersion < 0x80000000) {
			r = CreateProcessW(null, cast(wchar*) toUTF16z(process.dup), null, null, false, 0, null,
				workDir.length ? toUTF16z(workDir) : null, &setup, &info);
		} else {
			r = CreateProcessA(null, cast(char*) toStringz(process.dup), null, null, false, 0, null,
				workDir.length ? toStringz(workDir) : null, &setup, &info);
		}
		if (r) {
			CloseHandle(info.hThread);
			return true;
		}
		return false;
	}
	/// 環境変数を読込む。
	string getenv(string env) {
		return touni(std.process.getenv(env));
	}
} else {
	static assert (0, "Windows only.");
}

void main() {
	Application.enableVisualStyles;
	Application.autoCollect = false;
	Application.run(new List);
}

/// path以下の全フォルダからシナリオを検索し、
/// 見つかったシナリオの概要を返す。
/// loadingには随時読込中のシナリオ名が渡される。
/// errには随時発生したエラーのメッセージが渡される。
Summary[] find(string path, void delegate(string) loading, void delegate(string) err) {
	Summary[] list;
	try {
		void loadf(string sp) {
			try {
				scope fis = new BufferedFile(sp);
				scope (exit) fis.close;
				list ~= loadSummary(fis, path);
			} catch {
				if (err) err(sp ~ "の読込に失敗しました");
			}
		}
		if (path.isdir) {
			string sp = path.pjoin("Summary.wsm");
			if (sp.exists) {
				if (loading) loading(path);
				loadf(sp);
			} else {
				foreach (c; path.listdir) {
					c = path.pjoin(c);
					if (c.isdir || c.getExt.fnmatch("cab")) {
						list ~= find(c, loading, err);
					}
				}
			}
		} else if (canUncab && path.getExt.fnmatch("cab")) {
			auto tmp = getenv("TMP");
			if (!tmp || !tmp.exists) tmp = "";
			string sf = "Summary.wsm";
			int i = 1;
			while (tmp.pjoin(sf).exists) {
				i++;
				sf = format("Summary(%d).wsm", i);
			}
			if (loading) loading(path);
			if (uncab(path, tmp, (string file) {return file.basename.fnmatch("Summary.wsm") ? sf : "";})) {
				string sp = tmp.pjoin(sf);
				scope (exit) if (sp.exists) sp.fremove;
				loadf(sp);
			} else {
				if (err) err(path ~ "の読込に失敗しました");
			}
		}
	} catch {
		if (err) err(path ~ "の読込に失敗しました");
	}
	return list;
}

/// sを正しいファイル名にして返す。
string fname(string s) {
	char[] buf;
	char[4u] encBuf;
	foreach (c; s) {
		int i = std.string.indexOf(NO_FN_C, c);
		if (i >= 0) {
			buf ~= encBuf[0 .. encode(encBuf, NO_FN_CC[i])];
		} else {
			buf ~= c;
		}
	}
	return cast(immutable) buf;
}

/// fromからtoへファイルをコピーする。
/// fromがフォルダの場合、from以下の内容も全てコピーされる。
void deepcopy(string from, string to) {
	if (from.isdir) {
		if (!to.exists) to.mkdir;
		foreach (file; from.listdir) {
			deepcopy(from.pjoin(file), to.pjoin(file));
		}
	} else {
		from.fcopy(to);
	}
}

/// path以下の全ファイル・フォルダを削除する。
void delAll(string path) {
	if (path.exists) {
		if (path.isdir) {
			foreach (string f; path.dirEntries(SpanMode.depth)) {
				if (f.isdir) {
					f.rmdir;
				} else {
					f.fremove;
				}
			}
			path.rmdir;
		} else {
			path.fremove;
		}
	}
}

/// 整理の方式。
enum OrderType {
	AUTHOR, /// 作成者別。
	LEVEL /// 対象レベル別。
}

/// CWScViewerのメインForm。
class List : Form {
	private static class STag {
		Summary* s;
		alias s this;
		this (ref Summary s) {this.s = &s;}
	}

	private MenuItem menuOpen;
	private Button putInOrderExec;
	private class PutInOrder : Panel {
		private ComboBox orderType;
		private ComboBox putType;
		private CheckBox overwrite;
		this () {
			int w = 0, h = 0;
			void put(Size size) {
				w += size.width;
				h = max(h, size.height);
			}
			Label label(string text) {
				auto label = new Label;
				label.text = text;
				label.size = label.preferredSize;
				put(label.size);
				return label;
			}
			ComboBox combo(string[] items, int width) {
				auto combo = new ComboBox;
				combo.dropDownStyle = ComboBoxStyle.DROP_DOWN_LIST;
				combo.items.addRange(items);
				combo.width = width;
				put(combo.size);
				return combo;
			}
			Button button(string text, int width) {
				auto button = new Button;
				button.text = text;
				button.width = width;
				put(button.size);
				return button;
			}
			controls.add(label("これらのシナリオを"));
			controls.add(orderType = combo(["作成者", "対象レベル"], 85));
			controls.add(label("順に整理して"));
			controls.add(putType = combo(["コピー", "移動"], 60));
			controls.add(putInOrderExec = button("フォルダを選択して実行", 130));
			overwrite = new CheckBox;
			overwrite.text = "重複するシナリオがあったら上書きする";
			h += overwrite.height;
			overwrite.checked = false;
			controls.add(overwrite);
			putInOrderExec.click ~= (Object sencer, EventArgs args) {
				execute;
			};
			size = Size(w, h);
		}
		protected override void onLayout(LayoutEventArgs lea) {
			int x = 0;
			int bh = overwrite.height;
			int h = size.height;
			int th = h - bh;
			foreach (ctrl; controls) {
				auto size = ctrl.size;
				if (overwrite is ctrl) {
					ctrl.bounds = Rect(200, th, 200, size.height);
				} else {
					ctrl.bounds = Rect(x, (th - size.height) / 2, size.width, size.height);
					x += size.width;
					x += 5;
				}
			}
		}
		protected override void onVisibleChanged(EventArgs ea) {
			super.onVisibleChanged(ea);
			orderType.selectedIndex = 0;
			putType.selectedIndex = 0;
		}
		private void execute() {
			bool move = putType.selectedIndex == 1;
			OrderType order = orderType.selectedIndex == 0 ? OrderType.AUTHOR : OrderType.LEVEL;
			bool ow = overwrite.checked;
			auto sels = list.selectedItems;
			int c = sels.length ? sels.length : summs.length;
			auto dlg = new FolderBrowserDialog;
			dlg.description = format("%d件", c) ~ "のシナリオを"
				~ (order == OrderType.AUTHOR ? "作者" : "対象レベル") ~ "別に整理した結果の"
				~ (move ? "移動" : "コピー") ~ "先を選択(重複シナリオは"
				~ (ow ? "上書きされます" : "上書きされません") ~ ")";
			dlg.showNewFolderButton = true;
			dlg.selectedPath = getcwd;
			if (DialogResult.OK == dlg.showDialog) {
				auto path = dlg.selectedPath;
				if (path.listdir.length && DialogResult.YES != msgBox(this.outer,
						path ~ "は空ではありません。\n本当にここに整理結果を格納しますか？", "警告",
						MsgBoxButtons.YES_NO, MsgBoxIcon.QUESTION, MsgBoxDefaultButton.BUTTON2)) {
					return;
				}
				Summary*[] targs;
				if (sels.length) {
					foreach (ref itm; sels) {
						targs ~= (cast(STag) itm.tag).s;
					}
				} else {
					foreach (ref summ; summs) {
						targs ~= &summ;
					}
				}
				menuOpen.enabled = false;
				putInOrderExec.enabled = false;
				(new Thread(&(new Exec(path, move, order, ow, targs)).run)).start;
			}
		}
		private class Exec {
			private string path;
			private bool move;
			private OrderType order;
			private bool overwrite;
			private Summary*[] summs;
			this (string path, bool move, OrderType order, bool overwrite, Summary*[] summs) {
				this.path = path;
				this.move = move;
				this.order = order;
				this.overwrite = overwrite;
				this.summs = summs;
			}
			void run() {
				scope (exit) {
					this.outer.outer.invoke({
						menuOpen.enabled = true;
						putInOrderExec.enabled = true;
					});
				}
				void exec(ref Summary* summ, ref int count, int max) {
					if (!summ.sPath.exists) {
						warning(summ.sPath ~ "が見つかりません");
						return;
					}
					try {
						string dir;
						final switch (order) {
						case OrderType.AUTHOR: {
							dir = fname(summ.author);
							if (!dir.length) dir = "(noname)";
						} break;
						case OrderType.LEVEL: {
							dir = format("level_%02d-%02d", summ.levelMin, summ.levelMax);
						} break;
						}
						dir = path.pjoin(dir);
						if (!dir.exists) mkdir(dir);
						string to = dir.pjoin(summ.sPath.basename);
						string cs = format("%d / %d 件完了 ", count, max) ~ summ.name;
						count++;
						if (summ.sPath.fnmatch(to)) return;
						if (overwrite) {
							to.delAll;
						} else {
							int i = 1;
							while (to.exists) {
								i++;
								if (summ.sPath.isdir) {
									to = dir.pjoin(summ.sPath.basename) ~ format("(%d)", i);
								} else {
									to = (dir.pjoin(summ.sPath.basename.getName)
										~ format("(%d)", i)).addExt(summ.sPath.getExt);
								}
							}
						}
						if (move) {
							statusMsg(cs ~ "を移動中...");
							summ.sPath.rename(to);
						} else {
							statusMsg(cs ~ "をコピー中...");
							summ.sPath.deepcopy(to);
						}
						summ.sPath = to;
					} catch (Exception e) {
						warning(summ.sPath ~ "の" ~ (move ? "移動" : "コピー") ~ "に失敗しました\n" ~ e.msg);
					}
				}
				auto sels = list.selectedItems;
				int count = 0;
				foreach (ref summ; summs) {
					exec(summ, count, summs.length);
				}
				invoke({this.outer.outer.text = path ~ " - " ~ TITLE;});
				statusMsg("整理完了" ~ format("(%d件中%d件)", summs.length, count));
			}
		}
	}

	private static const TITLE = "シナリオ一覧";
	private Summary[] summs;
	this () {
		text = TITLE;
		setupList;
		setupMenu;
		controls.add(pio = new PutInOrder);
		controls.add(status = new StatusBar);
		bounds = Rect(0, 0, 460, 400);
		refresh;
	}
	private void open(Object sender, EventArgs args) {
		auto dlg = new FolderBrowserDialog;
		dlg.description = "シナリオが含まれるフォルダを選択";
		dlg.showNewFolderButton = false;
		dlg.selectedPath = getcwd;
		if (DialogResult.OK == dlg.showDialog) {
			chdir(dlg.selectedPath);
			load(dlg.selectedPath);
		}
	}
	private void load(string[] files ...) {
		if (files.length) {
			if (files.length == 1) {
				text = files[0] ~ " - " ~ TITLE;
			} else {
				text = format("%dつのフォルダ", files.length) ~ " - " ~ TITLE;
			}
			menuOpen.enabled = false;
			putInOrderExec.enabled = false;
			(new Thread({
				scope (exit) {
					invoke({menuOpen.enabled = true;});
				}
				summs.length = 0;
				foreach (file; files) {
					statusMsg(file ~ "を探索中...");
					summs ~= find(file, &loading, &warning);
				}
				invoke(&refresh);
			})).start;
		}
	}
	private void loading(string path) {
		invoke({status.text = path ~ "を読み込んでいます...";});
	}
	private void warning(string msg) {
		invoke({msgBox(this, msg, "警告", MsgBoxButtons.OK, MsgBoxIcon.WARNING);});
	}
	private void statusMsg(string msg) {
		invoke({status.text = msg;});
	}
	private void sort(ListView sender, ColumnClickEventArgs args) {
		static string st(string name) {
			return `
			list.beginUpdate;
			scope (exit) list.endUpdate;
			if (list.sorter == &sorter` ~ name ~ `) {
				list.sorter = &rev!(sorter` ~ name ~ `);
			} else {
				list.sorter = &sorter` ~ name ~ `;
			}`;
		}
		switch (args.column) {
		case COL_NAME: {
			mixin (st("Name"));
		} break;
		case COL_AUTHOR: {
			mixin (st("Author"));
		} break;
		case COL_LEVEL: {
			mixin (st("Level"));
		} break;
		case COL_DESC: {
			;
		} break;
		default: assert (0);
		}
	}
	private int rev(alias cmp)(ListViewItem i1, ListViewItem i2) {
		return -cmp(i1, i2);
	}
	private int sorterName(ListViewItem i1, ListViewItem i2) {
		return (cast(STag) i1.tag).name.icmp((cast(STag) i2.tag).name);
	}
	private int sorterAuthor(ListViewItem i1, ListViewItem i2) {
		return (cast(STag) i1.tag).author.icmp((cast(STag) i2.tag).author);
	}
	private int sorterLevel(ListViewItem i1, ListViewItem i2) {
		auto s1 = cast(STag) i1.tag;
		auto s2 = cast(STag) i2.tag;
		if (s1.levelMin < s2.levelMin) return -1;
		if (s1.levelMin > s2.levelMin) return 1;
		if (s1.levelMax < s2.levelMax) return -1;
		if (s1.levelMax > s2.levelMax) return 1;
		return 0;
	}
	private void refresh() {
		list.beginUpdate;
		scope (exit) list.endUpdate;
		list.clear;
		list.sorter = null;
		foreach (ref summ; summs) {
			list.addRow([
				summ.name,
				summ.author,
				format("%d ～ %d", summ.levelMin, summ.levelMax),
				summ.desc
			]).tag = new STag(summ);
		}
		status.text = format("%d件のシナリオ", summs.length);
		putInOrderExec.enabled = summs.length > 0;
	}
	private void exit(Object sender, EventArgs args) {
		Application.exit;
	}
	private static MenuItem menuItem(string text, void delegate(Object, EventArgs) click = null) {
		auto i = new MenuItem;
		i.text = text;
		if (click) i.click ~= click;
		return i;
	}
	private void setupMenu() {
		auto mm = new MainMenu;
		menu = mm;
		{
			auto mi = menuItem("ファイル(&F)");
			mm.menuItems.add(mi);
			mi.menuItems.add(menuOpen = menuItem("開く(&O)...\tCtrl+N", &open));
			addShortcut(Keys.CONTROL | Keys.N, &open);
			mi.menuItems.add(menuItem("-"));
			mi.menuItems.add(menuItem("終了(&Q)...\tAlt+F4", &exit));
			addShortcut(Keys.ALT | Keys.F4, &exit);
		}
	}
	private void selectSc() {
		foreach (itm; list.selectedItems) selectSc(*(cast(STag) itm.tag));
	}
	private void selectSc(in Summary summ) {
		if (summ.sPath.exists) {
			if (summ.sPath.isdir) {
				exec(`explorer "` ~ summ.sPath ~ `"`);
			} else {
				scUp(summ);
			}
		} else {
			warning(summ.sPath ~ "が見つかりません");
		}
	}
	private void scUp() {
		foreach (itm; list.selectedItems) scUp(*(cast(STag) itm.tag));
	}
	private void scUp(in Summary summ) {
		if (summ.sPath.exists) {
			exec(`explorer /select,"` ~ summ.sPath ~ `"`);
		} else {
			warning(summ.sPath ~ "が見つかりません");
		}
	}
	private StatusBar status;
	private PutInOrder pio;
	protected override void onLayout(LayoutEventArgs lea) {
		auto rect = clientRectangle;
		int y = rect.height;
		{
			auto size = status.size;
			y -= size.height;
			status.bounds = Rect(rect.x, y, rect.width, size.height);
		}
		{
			auto size = pio.size;
			y -= size.height;
			pio.bounds = Rect(rect.x, y, rect.width, size.height);
		}
		list.bounds = Rect(rect.x, rect.y, rect.width, y);
	}
	private ListView list;
	private void setupListMenu() {
		auto cm = new ContextMenu;
		list.contextMenu = cm;
		cm.menuItems.add(menuItem("フォルダを開く(&E)", (Object sender, EventArgs args) {selectSc;}));
		cm.menuItems.add(menuItem("-"));
		cm.menuItems.add(menuItem("一つ上のフォルダを開く(&U)", (Object sender, EventArgs args) {scUp;}));
	}
	private void listDragOver(Object sender, DragEventArgs ea) {
		if(ea.data.getDataPresent(DataFormats.fileDrop)) {
			ea.effect = ea.allowedEffect & DragDropEffects.LINK;
		}
	}
	private void listDragDrop(Object sender, DragEventArgs ea) {
		if(ea.data.getDataPresent(DataFormats.fileDrop) && (ea.allowedEffect & DragDropEffects.LINK)) {
			ea.effect = DragDropEffects.LINK;
			load(ea.data.getData(DataFormats.fileDrop).getStrings);
		}
	}
	private void setupList() {
		list = new ListView;
		list.view = View.DETAILS;
		list.allowDrop = true;
		list.dragOver ~= &listDragOver;
		list.dragDrop ~= &listDragDrop;
		setupListMenu;
		list.columnClick ~= &sort;
		list.keyPress ~= (Control sender, KeyPressEventArgs args) {
			if (args.keyCode == Keys.ENTER) {
				selectSc;
			}
		};
		list.doubleClick ~= (Control sender, EventArgs args) {selectSc;};
		ColumnHeader ch(string text, int width, HorizontalAlignment textAlign = HorizontalAlignment.LEFT) {
			auto ch = new ColumnHeader;
			ch.text = text;
			ch.width = width;
			ch.textAlign = textAlign;
			return ch;
		}
		list.columns.add(ch("シナリオ名", 120));
		list.columns.add(ch("作成者", 80));
		list.columns.add(ch("対象レベル", 70, HorizontalAlignment.CENTER));
		list.columns.add(ch("解説", 150));
		controls.add(list);
	}
	private static const COL_NAME = 0;
	private static const COL_AUTHOR = 1;
	private static const COL_LEVEL = 2;
	private static const COL_DESC = 3;
}
